@@ -52,6 +52,7 @@ typedef struct {
     char *ca_file;
     long options;
     char *sni_error;
+    long command;
 } state_t;
 
 static int ssl_index;
@@ -69,7 +70,7 @@ typedef unsigned __int32 uint32_t;
 #endif
 
 void __free(void *ptr, size_t size) {
-  enif_free(ptr);
+    enif_free(ptr);
 }
 
 #undef uthash_malloc
@@ -105,9 +106,6 @@ static int set_option_flag(const unsigned char *opt, size_t len, long *flag) {
 typedef struct {
     char *key;
     char *file;
-    time_t key_mtime;
-    time_t dh_mtime;
-    time_t ca_mtime;
     SSL_CTX *ssl_ctx;
     UT_hash_handle hh;
 } cert_info_t;
@@ -119,13 +117,37 @@ static ErlNifRWLock *certs_map_lock = NULL;
 static ErlNifRWLock *certfiles_map_lock = NULL;
 
 static void free_cert_info(cert_info_t *info) {
-  if (info) {
-    enif_free(info->key);
-    enif_free(info->file);
-    if (info->ssl_ctx)
-      SSL_CTX_free(info->ssl_ctx);
-    enif_free(info);
-  }
+    if (info) {
+        enif_free(info->key);
+        enif_free(info->file);
+        if (info->ssl_ctx)
+            SSL_CTX_free(info->ssl_ctx);
+        enif_free(info);
+    }
+}
+
+static void clear_certs_map() {
+    cert_info_t *info = NULL;
+    cert_info_t *tmp = NULL;
+
+    enif_rwlock_rwlock(certs_map_lock);
+    HASH_ITER(hh, certs_map, info, tmp) {
+        HASH_DEL(certs_map, info);
+        free_cert_info(info);
+    }
+    enif_rwlock_rwunlock(certs_map_lock);
+}
+
+static void clear_certfiles_map() {
+    cert_info_t *info = NULL;
+    cert_info_t *tmp = NULL;
+
+    enif_rwlock_rwlock(certfiles_map_lock);
+    HASH_ITER(hh, certfiles_map, info, tmp) {
+        HASH_DEL(certfiles_map, info);
+        free_cert_info(info);
+    }
+    enif_rwlock_rwunlock(certfiles_map_lock);
 }
 
 static state_t *init_tls_state() {
@@ -134,8 +156,8 @@ static state_t *init_tls_state() {
     memset(state, 0, sizeof(state_t));
     state->mtx = enif_mutex_create("");
     if (!state->mtx) {
-      enif_release_resource(state);
-      return NULL;
+        enif_release_resource(state);
+        return NULL;
     }
     state->valid = 1;
     return state;
@@ -152,8 +174,8 @@ static void destroy_tls_state(ErlNifEnv *env, void *data) {
             enif_free(state->send_buffer);
         if (state->send_buffer2)
             enif_free(state->send_buffer2);
-	if (state->cert_file)
-	    enif_free(state->cert_file);
+        if (state->cert_file)
+            enif_free(state->cert_file);
         memset(state, 0, sizeof(state_t));
     }
 }
@@ -200,21 +222,9 @@ static int load(ErlNifEnv *env, void **priv, ERL_NIF_TERM load_info) {
 
 static void unload(ErlNifEnv *env, void *priv) {
     int i;
-    cert_info_t *info = NULL;
-    cert_info_t *tmp = NULL;
 
-    enif_rwlock_rwlock(certs_map_lock);
-    HASH_ITER(hh, certs_map, info, tmp) {
-      HASH_DEL(certs_map, info);
-      free_cert_info(info);
-    }
-    enif_rwlock_rwunlock(certs_map_lock);
-    enif_rwlock_rwlock(certfiles_map_lock);
-    HASH_ITER(hh, certfiles_map, info, tmp) {
-      HASH_DEL(certfiles_map, info);
-      free_cert_info(info);
-    }
-    enif_rwlock_rwunlock(certfiles_map_lock);
+    clear_certs_map();
+    clear_certfiles_map();
     enif_rwlock_destroy(certs_map_lock);
     enif_rwlock_destroy(certfiles_map_lock);
     certs_map = NULL;
@@ -227,23 +237,6 @@ static void unload(ErlNifEnv *env, void *priv) {
     mtx_buf = NULL;
 }
 
-static int is_modified(char *file, time_t *known_mtime) {
-    struct stat file_stat;
-
-    if (file == NULL) {
-        return 0;
-    } else if (stat(file, &file_stat)) {
-        *known_mtime = 0;
-        return 1;
-    } else {
-        if (*known_mtime != file_stat.st_mtime) {
-            *known_mtime = file_stat.st_mtime;
-            return 1;
-        } else
-            return 0;
-    }
-}
-
 static int verify_callback(int preverify_ok, X509_STORE_CTX *ctx) {
     return 1;
 }
@@ -280,72 +273,72 @@ static void setup_ecdh(SSL_CTX *ctx) {
   2048-bit MODP Group with 256-bit Prime Order Subgroup (RFC5114)
 */
 static unsigned char dh2048_p[] = {
-  0x87, 0xA8, 0xE6, 0x1D, 0xB4, 0xB6, 0x66, 0x3C,
-  0xFF, 0xBB, 0xD1, 0x9C, 0x65, 0x19, 0x59, 0x99,
-  0x8C, 0xEE, 0xF6, 0x08, 0x66, 0x0D, 0xD0, 0xF2,
-  0x5D, 0x2C, 0xEE, 0xD4, 0x43, 0x5E, 0x3B, 0x00,
-  0xE0, 0x0D, 0xF8, 0xF1, 0xD6, 0x19, 0x57, 0xD4,
-  0xFA, 0xF7, 0xDF, 0x45, 0x61, 0xB2, 0xAA, 0x30,
-  0x16, 0xC3, 0xD9, 0x11, 0x34, 0x09, 0x6F, 0xAA,
-  0x3B, 0xF4, 0x29, 0x6D, 0x83, 0x0E, 0x9A, 0x7C,
-  0x20, 0x9E, 0x0C, 0x64, 0x97, 0x51, 0x7A, 0xBD,
-  0x5A, 0x8A, 0x9D, 0x30, 0x6B, 0xCF, 0x67, 0xED,
-  0x91, 0xF9, 0xE6, 0x72, 0x5B, 0x47, 0x58, 0xC0,
-  0x22, 0xE0, 0xB1, 0xEF, 0x42, 0x75, 0xBF, 0x7B,
-  0x6C, 0x5B, 0xFC, 0x11, 0xD4, 0x5F, 0x90, 0x88,
-  0xB9, 0x41, 0xF5, 0x4E, 0xB1, 0xE5, 0x9B, 0xB8,
-  0xBC, 0x39, 0xA0, 0xBF, 0x12, 0x30, 0x7F, 0x5C,
-  0x4F, 0xDB, 0x70, 0xC5, 0x81, 0xB2, 0x3F, 0x76,
-  0xB6, 0x3A, 0xCA, 0xE1, 0xCA, 0xA6, 0xB7, 0x90,
-  0x2D, 0x52, 0x52, 0x67, 0x35, 0x48, 0x8A, 0x0E,
-  0xF1, 0x3C, 0x6D, 0x9A, 0x51, 0xBF, 0xA4, 0xAB,
-  0x3A, 0xD8, 0x34, 0x77, 0x96, 0x52, 0x4D, 0x8E,
-  0xF6, 0xA1, 0x67, 0xB5, 0xA4, 0x18, 0x25, 0xD9,
-  0x67, 0xE1, 0x44, 0xE5, 0x14, 0x05, 0x64, 0x25,
-  0x1C, 0xCA, 0xCB, 0x83, 0xE6, 0xB4, 0x86, 0xF6,
-  0xB3, 0xCA, 0x3F, 0x79, 0x71, 0x50, 0x60, 0x26,
-  0xC0, 0xB8, 0x57, 0xF6, 0x89, 0x96, 0x28, 0x56,
-  0xDE, 0xD4, 0x01, 0x0A, 0xBD, 0x0B, 0xE6, 0x21,
-  0xC3, 0xA3, 0x96, 0x0A, 0x54, 0xE7, 0x10, 0xC3,
-  0x75, 0xF2, 0x63, 0x75, 0xD7, 0x01, 0x41, 0x03,
-  0xA4, 0xB5, 0x43, 0x30, 0xC1, 0x98, 0xAF, 0x12,
-  0x61, 0x16, 0xD2, 0x27, 0x6E, 0x11, 0x71, 0x5F,
-  0x69, 0x38, 0x77, 0xFA, 0xD7, 0xEF, 0x09, 0xCA,
-  0xDB, 0x09, 0x4A, 0xE9, 0x1E, 0x1A, 0x15, 0x97,
+        0x87, 0xA8, 0xE6, 0x1D, 0xB4, 0xB6, 0x66, 0x3C,
+        0xFF, 0xBB, 0xD1, 0x9C, 0x65, 0x19, 0x59, 0x99,
+        0x8C, 0xEE, 0xF6, 0x08, 0x66, 0x0D, 0xD0, 0xF2,
+        0x5D, 0x2C, 0xEE, 0xD4, 0x43, 0x5E, 0x3B, 0x00,
+        0xE0, 0x0D, 0xF8, 0xF1, 0xD6, 0x19, 0x57, 0xD4,
+        0xFA, 0xF7, 0xDF, 0x45, 0x61, 0xB2, 0xAA, 0x30,
+        0x16, 0xC3, 0xD9, 0x11, 0x34, 0x09, 0x6F, 0xAA,
+        0x3B, 0xF4, 0x29, 0x6D, 0x83, 0x0E, 0x9A, 0x7C,
+        0x20, 0x9E, 0x0C, 0x64, 0x97, 0x51, 0x7A, 0xBD,
+        0x5A, 0x8A, 0x9D, 0x30, 0x6B, 0xCF, 0x67, 0xED,
+        0x91, 0xF9, 0xE6, 0x72, 0x5B, 0x47, 0x58, 0xC0,
+        0x22, 0xE0, 0xB1, 0xEF, 0x42, 0x75, 0xBF, 0x7B,
+        0x6C, 0x5B, 0xFC, 0x11, 0xD4, 0x5F, 0x90, 0x88,
+        0xB9, 0x41, 0xF5, 0x4E, 0xB1, 0xE5, 0x9B, 0xB8,
+        0xBC, 0x39, 0xA0, 0xBF, 0x12, 0x30, 0x7F, 0x5C,
+        0x4F, 0xDB, 0x70, 0xC5, 0x81, 0xB2, 0x3F, 0x76,
+        0xB6, 0x3A, 0xCA, 0xE1, 0xCA, 0xA6, 0xB7, 0x90,
+        0x2D, 0x52, 0x52, 0x67, 0x35, 0x48, 0x8A, 0x0E,
+        0xF1, 0x3C, 0x6D, 0x9A, 0x51, 0xBF, 0xA4, 0xAB,
+        0x3A, 0xD8, 0x34, 0x77, 0x96, 0x52, 0x4D, 0x8E,
+        0xF6, 0xA1, 0x67, 0xB5, 0xA4, 0x18, 0x25, 0xD9,
+        0x67, 0xE1, 0x44, 0xE5, 0x14, 0x05, 0x64, 0x25,
+        0x1C, 0xCA, 0xCB, 0x83, 0xE6, 0xB4, 0x86, 0xF6,
+        0xB3, 0xCA, 0x3F, 0x79, 0x71, 0x50, 0x60, 0x26,
+        0xC0, 0xB8, 0x57, 0xF6, 0x89, 0x96, 0x28, 0x56,
+        0xDE, 0xD4, 0x01, 0x0A, 0xBD, 0x0B, 0xE6, 0x21,
+        0xC3, 0xA3, 0x96, 0x0A, 0x54, 0xE7, 0x10, 0xC3,
+        0x75, 0xF2, 0x63, 0x75, 0xD7, 0x01, 0x41, 0x03,
+        0xA4, 0xB5, 0x43, 0x30, 0xC1, 0x98, 0xAF, 0x12,
+        0x61, 0x16, 0xD2, 0x27, 0x6E, 0x11, 0x71, 0x5F,
+        0x69, 0x38, 0x77, 0xFA, 0xD7, 0xEF, 0x09, 0xCA,
+        0xDB, 0x09, 0x4A, 0xE9, 0x1E, 0x1A, 0x15, 0x97,
 };
 static unsigned char dh2048_g[] = {
-  0x3F, 0xB3, 0x2C, 0x9B, 0x73, 0x13, 0x4D, 0x0B,
-  0x2E, 0x77, 0x50, 0x66, 0x60, 0xED, 0xBD, 0x48,
-  0x4C, 0xA7, 0xB1, 0x8F, 0x21, 0xEF, 0x20, 0x54,
-  0x07, 0xF4, 0x79, 0x3A, 0x1A, 0x0B, 0xA1, 0x25,
-  0x10, 0xDB, 0xC1, 0x50, 0x77, 0xBE, 0x46, 0x3F,
-  0xFF, 0x4F, 0xED, 0x4A, 0xAC, 0x0B, 0xB5, 0x55,
-  0xBE, 0x3A, 0x6C, 0x1B, 0x0C, 0x6B, 0x47, 0xB1,
-  0xBC, 0x37, 0x73, 0xBF, 0x7E, 0x8C, 0x6F, 0x62,
-  0x90, 0x12, 0x28, 0xF8, 0xC2, 0x8C, 0xBB, 0x18,
-  0xA5, 0x5A, 0xE3, 0x13, 0x41, 0x00, 0x0A, 0x65,
-  0x01, 0x96, 0xF9, 0x31, 0xC7, 0x7A, 0x57, 0xF2,
-  0xDD, 0xF4, 0x63, 0xE5, 0xE9, 0xEC, 0x14, 0x4B,
-  0x77, 0x7D, 0xE6, 0x2A, 0xAA, 0xB8, 0xA8, 0x62,
-  0x8A, 0xC3, 0x76, 0xD2, 0x82, 0xD6, 0xED, 0x38,
-  0x64, 0xE6, 0x79, 0x82, 0x42, 0x8E, 0xBC, 0x83,
-  0x1D, 0x14, 0x34, 0x8F, 0x6F, 0x2F, 0x91, 0x93,
-  0xB5, 0x04, 0x5A, 0xF2, 0x76, 0x71, 0x64, 0xE1,
-  0xDF, 0xC9, 0x67, 0xC1, 0xFB, 0x3F, 0x2E, 0x55,
-  0xA4, 0xBD, 0x1B, 0xFF, 0xE8, 0x3B, 0x9C, 0x80,
-  0xD0, 0x52, 0xB9, 0x85, 0xD1, 0x82, 0xEA, 0x0A,
-  0xDB, 0x2A, 0x3B, 0x73, 0x13, 0xD3, 0xFE, 0x14,
-  0xC8, 0x48, 0x4B, 0x1E, 0x05, 0x25, 0x88, 0xB9,
-  0xB7, 0xD2, 0xBB, 0xD2, 0xDF, 0x01, 0x61, 0x99,
-  0xEC, 0xD0, 0x6E, 0x15, 0x57, 0xCD, 0x09, 0x15,
-  0xB3, 0x35, 0x3B, 0xBB, 0x64, 0xE0, 0xEC, 0x37,
-  0x7F, 0xD0, 0x28, 0x37, 0x0D, 0xF9, 0x2B, 0x52,
-  0xC7, 0x89, 0x14, 0x28, 0xCD, 0xC6, 0x7E, 0xB6,
-  0x18, 0x4B, 0x52, 0x3D, 0x1D, 0xB2, 0x46, 0xC3,
-  0x2F, 0x63, 0x07, 0x84, 0x90, 0xF0, 0x0E, 0xF8,
-  0xD6, 0x47, 0xD1, 0x48, 0xD4, 0x79, 0x54, 0x51,
-  0x5E, 0x23, 0x27, 0xCF, 0xEF, 0x98, 0xC5, 0x82,
-  0x66, 0x4B, 0x4C, 0x0F, 0x6C, 0xC4, 0x16, 0x59,
+        0x3F, 0xB3, 0x2C, 0x9B, 0x73, 0x13, 0x4D, 0x0B,
+        0x2E, 0x77, 0x50, 0x66, 0x60, 0xED, 0xBD, 0x48,
+        0x4C, 0xA7, 0xB1, 0x8F, 0x21, 0xEF, 0x20, 0x54,
+        0x07, 0xF4, 0x79, 0x3A, 0x1A, 0x0B, 0xA1, 0x25,
+        0x10, 0xDB, 0xC1, 0x50, 0x77, 0xBE, 0x46, 0x3F,
+        0xFF, 0x4F, 0xED, 0x4A, 0xAC, 0x0B, 0xB5, 0x55,
+        0xBE, 0x3A, 0x6C, 0x1B, 0x0C, 0x6B, 0x47, 0xB1,
+        0xBC, 0x37, 0x73, 0xBF, 0x7E, 0x8C, 0x6F, 0x62,
+        0x90, 0x12, 0x28, 0xF8, 0xC2, 0x8C, 0xBB, 0x18,
+        0xA5, 0x5A, 0xE3, 0x13, 0x41, 0x00, 0x0A, 0x65,
+        0x01, 0x96, 0xF9, 0x31, 0xC7, 0x7A, 0x57, 0xF2,
+        0xDD, 0xF4, 0x63, 0xE5, 0xE9, 0xEC, 0x14, 0x4B,
+        0x77, 0x7D, 0xE6, 0x2A, 0xAA, 0xB8, 0xA8, 0x62,
+        0x8A, 0xC3, 0x76, 0xD2, 0x82, 0xD6, 0xED, 0x38,
+        0x64, 0xE6, 0x79, 0x82, 0x42, 0x8E, 0xBC, 0x83,
+        0x1D, 0x14, 0x34, 0x8F, 0x6F, 0x2F, 0x91, 0x93,
+        0xB5, 0x04, 0x5A, 0xF2, 0x76, 0x71, 0x64, 0xE1,
+        0xDF, 0xC9, 0x67, 0xC1, 0xFB, 0x3F, 0x2E, 0x55,
+        0xA4, 0xBD, 0x1B, 0xFF, 0xE8, 0x3B, 0x9C, 0x80,
+        0xD0, 0x52, 0xB9, 0x85, 0xD1, 0x82, 0xEA, 0x0A,
+        0xDB, 0x2A, 0x3B, 0x73, 0x13, 0xD3, 0xFE, 0x14,
+        0xC8, 0x48, 0x4B, 0x1E, 0x05, 0x25, 0x88, 0xB9,
+        0xB7, 0xD2, 0xBB, 0xD2, 0xDF, 0x01, 0x61, 0x99,
+        0xEC, 0xD0, 0x6E, 0x15, 0x57, 0xCD, 0x09, 0x15,
+        0xB3, 0x35, 0x3B, 0xBB, 0x64, 0xE0, 0xEC, 0x37,
+        0x7F, 0xD0, 0x28, 0x37, 0x0D, 0xF9, 0x2B, 0x52,
+        0xC7, 0x89, 0x14, 0x28, 0xCD, 0xC6, 0x7E, 0xB6,
+        0x18, 0x4B, 0x52, 0x3D, 0x1D, 0xB2, 0x46, 0xC3,
+        0x2F, 0x63, 0x07, 0x84, 0x90, 0xF0, 0x0E, 0xF8,
+        0xD6, 0x47, 0xD1, 0x48, 0xD4, 0x79, 0x54, 0x51,
+        0x5E, 0x23, 0x27, 0xCF, 0xEF, 0x98, 0xC5, 0x82,
+        0x66, 0x4B, 0x4C, 0x0F, 0x6C, 0xC4, 0x16, 0x59,
 };
 
 static int setup_dh(SSL_CTX *ctx, char *dh_file) {
@@ -401,61 +394,61 @@ static void ssl_info_callback(const SSL *s, int where, int ret) {
 static char *create_ssl_for_cert(char *, state_t *);
 
 static cert_info_t *lookup_certfile(const char *domain) {
-  cert_info_t *ret = NULL;
-  cert_info_t *info = NULL;
+    cert_info_t *ret = NULL;
+    cert_info_t *info = NULL;
 
-  if (domain) {
-    size_t len = strlen(domain);
-    if (len) {
-      char name[len+1];
-      name[len] = 0;
-      size_t i = 0;
-      for (i=0; i<len; i++)
-	name[i] = tolower(domain[i]);
-      HASH_FIND_STR(certfiles_map, name, info);
-      if (info && info->file)
-	ret = info;
-      else {
-	/* Replace the first domain part with '*' and retry */
-	char *dot = strchr(name, '.');
-	if (dot != NULL && name[0] != '.') {
-	  char *glob = dot - 1;
-	  glob[0] = '*';
-	  HASH_FIND_STR(certfiles_map, glob, info);
-	  if (info && info->file)
-	    ret = info;
-	}
-      }
+    if (domain) {
+        size_t len = strlen(domain);
+        if (len) {
+            char name[len + 1];
+            name[len] = 0;
+            size_t i = 0;
+            for (i = 0; i < len; i++)
+                name[i] = tolower(domain[i]);
+            HASH_FIND_STR(certfiles_map, name, info);
+            if (info && info->file)
+                ret = info;
+            else {
+                /* Replace the first domain part with '*' and retry */
+                char *dot = strchr(name, '.');
+                if (dot != NULL && name[0] != '.') {
+                    char *glob = dot - 1;
+                    glob[0] = '*';
+                    HASH_FIND_STR(certfiles_map, glob, info);
+                    if (info && info->file)
+                        ret = info;
+                }
+            }
+        }
     }
-  }
-  return ret;
+    return ret;
 }
 
 static int ssl_sni_callback(const SSL *s, int *foo, void *data) {
-  cert_info_t *info = NULL;
-  char *err_str = NULL;
-  const char *servername = NULL;
-  int ret = SSL_TLSEXT_ERR_OK;
-  state_t *state = (state_t *) SSL_get_ex_data(s, ssl_index);
+    cert_info_t *info = NULL;
+    char *err_str = NULL;
+    const char *servername = NULL;
+    int ret = SSL_TLSEXT_ERR_OK;
+    state_t *state = (state_t *) SSL_get_ex_data(s, ssl_index);
 
-  servername = SSL_get_servername(s, TLSEXT_NAMETYPE_host_name);
-  enif_rwlock_rlock(certfiles_map_lock);
-  info = lookup_certfile(servername);
-  if (info) {
-    if (strcmp(info->file, state->cert_file))
-      err_str = create_ssl_for_cert(info->file, state);
-    if (err_str) {
-      state->sni_error = err_str;
-      ret = SSL_TLSEXT_ERR_ALERT_FATAL;
+    servername = SSL_get_servername(s, TLSEXT_NAMETYPE_host_name);
+    enif_rwlock_rlock(certfiles_map_lock);
+    info = lookup_certfile(servername);
+    if (info) {
+        if (strcmp(info->file, state->cert_file))
+            err_str = create_ssl_for_cert(info->file, state);
+        if (err_str) {
+            state->sni_error = err_str;
+            ret = SSL_TLSEXT_ERR_ALERT_FATAL;
+        }
+    } else if (strlen(state->cert_file) == 0) {
+        state->sni_error =
+                "Failed to find a certificate matching the domain in SNI extension";
+        ret = SSL_TLSEXT_ERR_ALERT_FATAL;
     }
-  } else if (strlen(state->cert_file) == 0) {
-    state->sni_error =
-      "Failed to find a certificate matching the domain in SNI extension";
-    ret = SSL_TLSEXT_ERR_ALERT_FATAL;
-  }
-  enif_rwlock_runlock(certfiles_map_lock);
+    enif_rwlock_runlock(certfiles_map_lock);
 
-  return ret;
+    return ret;
 }
 
 #define ERR_T(T) enif_make_tuple2(env, enif_make_atom(env, "error"), T)
@@ -488,79 +481,87 @@ static ERL_NIF_TERM ssl_error(ErlNifEnv *env, const char *errstr) {
 }
 
 static SSL_CTX *create_new_ctx(char *cert_file, char *ciphers,
-			       char *dh_file, char *ca_file,
-			       char **err_str) {
-  int res = 0;
+                               char *dh_file, char *ca_file,
+                               unsigned int command,
+                               char **err_str) {
+    long verifyopts;
+    int res = 0;
 
-  SSL_CTX *ctx = SSL_CTX_new(SSLv23_method());
-  if (!ctx) {
-    *err_str = "SSL_CTX_new failed";
-    return NULL;
-  }
-  if (cert_file) {
-    res = SSL_CTX_use_certificate_chain_file(ctx, cert_file);
-    if (res <= 0) {
-      SSL_CTX_free(ctx);
-      *err_str = "SSL_CTX_use_certificate_file failed";
-      return NULL;
+    SSL_CTX *ctx = SSL_CTX_new(SSLv23_method());
+    if (!ctx) {
+        *err_str = "SSL_CTX_new failed";
+        return NULL;
     }
-    res = SSL_CTX_use_PrivateKey_file(ctx, cert_file, SSL_FILETYPE_PEM);
-    if (res <= 0) {
-      SSL_CTX_free(ctx);
-      *err_str = "SSL_CTX_use_PrivateKey_file failed";
-      return NULL;
+    if (cert_file) {
+        res = SSL_CTX_use_certificate_chain_file(ctx, cert_file);
+        if (res <= 0) {
+            SSL_CTX_free(ctx);
+            *err_str = "SSL_CTX_use_certificate_file failed";
+            return NULL;
+        }
+        res = SSL_CTX_use_PrivateKey_file(ctx, cert_file, SSL_FILETYPE_PEM);
+        if (res <= 0) {
+            SSL_CTX_free(ctx);
+            *err_str = "SSL_CTX_use_PrivateKey_file failed";
+            return NULL;
+        }
+        res = SSL_CTX_check_private_key(ctx);
+        if (res <= 0) {
+            SSL_CTX_free(ctx);
+            *err_str = "SSL_CTX_check_private_key failed";
+            return NULL;
+        }
     }
-    res = SSL_CTX_check_private_key(ctx);
-    if (res <= 0) {
-      SSL_CTX_free(ctx);
-      *err_str = "SSL_CTX_check_private_key failed";
-      return NULL;
-    }
-  }
 
-  SSL_CTX_set_tlsext_servername_callback(ctx, &ssl_sni_callback);
+    if (command == SET_CERTIFICATE_FILE_ACCEPT) {
+        SSL_CTX_set_tlsext_servername_callback(ctx, &ssl_sni_callback);
+        verifyopts = SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE;
+        if (ca_file) {
+            SSL_CTX_set_client_CA_list(ctx, SSL_load_client_CA_file(ca_file));
+        }
+    } else {
+        verifyopts = SSL_VERIFY_PEER;
+    }
 
-  if (ciphers[0] == 0)
-    SSL_CTX_set_cipher_list(ctx, CIPHERS);
-  else
-    SSL_CTX_set_cipher_list(ctx, ciphers);
+    if (ciphers[0] == 0)
+        SSL_CTX_set_cipher_list(ctx, CIPHERS);
+    else
+        SSL_CTX_set_cipher_list(ctx, ciphers);
 
 #ifndef OPENSSL_NO_ECDH
-  setup_ecdh(ctx);
+    setup_ecdh(ctx);
 #endif
 #ifndef OPENSSL_NO_DH
-  res = setup_dh(ctx, dh_file);
-  if (res <= 0) {
-    SSL_CTX_free(ctx);
-    *err_str = "Setting DH parameters failed";
-    return NULL;
-  }
+    res = setup_dh(ctx, dh_file);
+    if (res <= 0) {
+        SSL_CTX_free(ctx);
+        *err_str = "Setting DH parameters failed";
+        return NULL;
+    }
 #endif
 
-  SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);
-  if (ca_file)
-    SSL_CTX_load_verify_locations(ctx, ca_file, NULL);
-  else
-    SSL_CTX_set_default_verify_paths(ctx);
+    SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_OFF);
+    if (ca_file)
+        SSL_CTX_load_verify_locations(ctx, ca_file, NULL);
+    else
+        SSL_CTX_set_default_verify_paths(ctx);
 
 #ifdef SSL_MODE_RELEASE_BUFFERS
-  SSL_CTX_set_mode(ctx, SSL_MODE_RELEASE_BUFFERS);
+    SSL_CTX_set_mode(ctx, SSL_MODE_RELEASE_BUFFERS);
 #endif
-  SSL_CTX_set_verify(ctx,
-		     SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE,
-		     verify_callback);
+    SSL_CTX_set_verify(ctx, verifyopts, verify_callback);
 
-  SSL_CTX_set_info_callback(ctx, &ssl_info_callback);
+    SSL_CTX_set_info_callback(ctx, &ssl_info_callback);
 
-  *err_str = NULL;
-  return ctx;
+    *err_str = NULL;
+    return ctx;
 }
 
 static void set_ctx(state_t *state, SSL_CTX *ctx) {
-  if (state->ssl)
-    SSL_set_SSL_CTX(state->ssl, ctx);
-  else
-    state->ssl = SSL_new(ctx);
+    if (state->ssl)
+        SSL_set_SSL_CTX(state->ssl, ctx);
+    else
+        state->ssl = SSL_new(ctx);
 }
 
 static char *create_ssl_for_cert(char *cert_file, state_t *state) {
@@ -568,14 +569,15 @@ static char *create_ssl_for_cert(char *cert_file, state_t *state) {
     char *dh_file = state->dh_file;
     char *ca_file = state->ca_file;
     long options = state->options;
+    unsigned int command = state->command;
 
     char *ret = NULL;
     cert_info_t *info = NULL;
     cert_info_t *new_info = NULL;
     cert_info_t *old_info = NULL;
     size_t key_size =
-      strlen(cert_file) + strlen(ciphers) + 8 +
-      strlen(dh_file) + strlen(ca_file) + 1;
+            strlen(cert_file) + strlen(ciphers) + 8 +
+            strlen(dh_file) + strlen(ca_file) + 1;
     char key[key_size];
     sprintf(key, "%s%s%08lx%s%s", cert_file, ciphers,
             options, dh_file, ca_file);
@@ -584,50 +586,40 @@ static char *create_ssl_for_cert(char *cert_file, state_t *state) {
 
     HASH_FIND_STR(certs_map, key, info);
 
-    time_t key_mtime = info ? info->key_mtime : 0;
-    time_t dh_mtime = info ? info->dh_mtime : 0;
-    time_t ca_mtime = info ? info->ca_mtime : 0;
-
     if (strlen(cert_file) == 0) cert_file = NULL;
     if (strlen(dh_file) == 0) dh_file = NULL;
     if (strlen(ca_file) == 0) ca_file = NULL;
 
-    if (is_modified(cert_file, &key_mtime) ||
-        is_modified(dh_file, &dh_mtime) ||
-        is_modified(ca_file, &ca_mtime) ||
-        info == NULL) {
+    if (info == NULL) {
         enif_rwlock_runlock(certs_map_lock);
 
         enif_rwlock_rwlock(certs_map_lock);
-	SSL_CTX *ctx = create_new_ctx(cert_file, ciphers, dh_file, ca_file, &ret);
-	if (ret == NULL) {
-	  new_info = enif_alloc(sizeof(cert_info_t));
-	  if (new_info) {
-	    memset(new_info, 0, sizeof(cert_info_t));
-	    new_info->key = enif_alloc(key_size);
-	    if (new_info->key) {
-	      memcpy(new_info->key, key, key_size);
-	      new_info->key_mtime = key_mtime;
-	      new_info->dh_mtime = dh_mtime;
-	      new_info->ca_mtime = ca_mtime;
-	      new_info->ssl_ctx = ctx;
-	      HASH_REPLACE_STR(certs_map, key, new_info, old_info);
-	      free_cert_info(old_info);
-	      set_ctx(state, ctx);
-	    } else {
-	      enif_free(new_info);
-	      SSL_CTX_free(ctx);
-	      ret = "Memory allocation failed";
-	    }
-	  } else {
-	    SSL_CTX_free(ctx);
-	    ret = "Memory allocation failed";
-	  }
-	}
-	enif_rwlock_rwunlock(certs_map_lock);
+        SSL_CTX *ctx = create_new_ctx(cert_file, ciphers, dh_file, ca_file, command, &ret);
+        if (ret == NULL) {
+            new_info = enif_alloc(sizeof(cert_info_t));
+            if (new_info) {
+                memset(new_info, 0, sizeof(cert_info_t));
+                new_info->key = enif_alloc(key_size);
+                if (new_info->key) {
+                    memcpy(new_info->key, key, key_size);
+                    new_info->ssl_ctx = ctx;
+                    HASH_REPLACE_STR(certs_map, key, new_info, old_info);
+                    free_cert_info(old_info);
+                    set_ctx(state, ctx);
+                } else {
+                    enif_free(new_info);
+                    SSL_CTX_free(ctx);
+                    ret = "Memory allocation failed";
+                }
+            } else {
+                SSL_CTX_free(ctx);
+                ret = "Memory allocation failed";
+            }
+        }
+        enif_rwlock_rwunlock(certs_map_lock);
     } else {
-      set_ctx(state, info->ssl_ctx);
-      enif_rwlock_runlock(certs_map_lock);
+        set_ctx(state, info->ssl_ctx);
+        enif_rwlock_runlock(certs_map_lock);
     }
     return ret;
 }
@@ -673,11 +665,11 @@ static ERL_NIF_TERM open_nif(ErlNifEnv *env, int argc,
 
     command = flags & 0xffff;
     if (protocol_options_bin.size) {
-      po_len_left = protocol_options_bin.size;
-      po = protocol_options_bin.data;
+        po_len_left = protocol_options_bin.size;
+        po = protocol_options_bin.data;
     } else {
-      po = (unsigned char *) PROTOCOL_OPTIONS;
-      po_len_left = strlen((char *) po);
+        po = (unsigned char *) PROTOCOL_OPTIONS;
+        po_len_left = strlen((char *) po);
     }
 
     while (po_len_left) {
@@ -696,19 +688,20 @@ static ERL_NIF_TERM open_nif(ErlNifEnv *env, int argc,
     if (!state) return ERR_T(enif_make_atom(env, "enomem"));
 
     state->cert_file = enif_alloc(certfile_bin.size + 1 +
-				  ciphers_bin.size + 1 +
-				  dhfile_bin.size + 1 +
-				  cafile_bin.size + 1 +
-				  sni_bin.size + 1);
+                                  ciphers_bin.size + 1 +
+                                  dhfile_bin.size + 1 +
+                                  cafile_bin.size + 1 +
+                                  sni_bin.size + 1);
     if (!state->cert_file) {
-      enif_release_resource(state);
-      return ERR_T(enif_make_atom(env, "enomem"));
+        enif_release_resource(state);
+        return ERR_T(enif_make_atom(env, "enomem"));
     }
     state->ciphers = state->cert_file + certfile_bin.size + 1;
     state->dh_file = state->ciphers + ciphers_bin.size + 1;
     state->ca_file = state->dh_file + dhfile_bin.size + 1;
     sni = state->ca_file + cafile_bin.size + 1;
     state->options = options;
+    state->command = command;
 
     memcpy(state->cert_file, certfile_bin.data, certfile_bin.size);
     state->cert_file[certfile_bin.size] = 0;
@@ -723,13 +716,13 @@ static ERL_NIF_TERM open_nif(ErlNifEnv *env, int argc,
 
     char *err_str = create_ssl_for_cert(state->cert_file, state);
     if (err_str) {
-      enif_release_resource(state);
-      return ssl_error(env, err_str);
+        enif_release_resource(state);
+        return ssl_error(env, err_str);
     }
 
     if (!state->ssl) {
-      enif_release_resource(state);
-      return ssl_error(env, "SSL_new failed");
+        enif_release_resource(state);
+        return ssl_error(env, "SSL_new failed");
     }
 
     if (flags & VERIFY_NONE)
@@ -758,11 +751,11 @@ static ERL_NIF_TERM open_nif(ErlNifEnv *env, int argc,
 
         SSL_set_options(state->ssl, options);
 
-	if (strlen(sni)>0) SSL_set_tlsext_host_name(state->ssl, sni);
+        if (strlen(sni) > 0) SSL_set_tlsext_host_name(state->ssl, sni);
 
 #if OPENSSL_VERSION_NUMBER >= 0x10002000L
-	if (alpn_bin.size)
-	  SSL_set_alpn_protos(state->ssl, alpn_bin.data, alpn_bin.size);
+        if (alpn_bin.size)
+            SSL_set_alpn_protos(state->ssl, alpn_bin.data, alpn_bin.size);
 #endif
 
         SSL_set_connect_state(state->ssl);
@@ -1002,18 +995,18 @@ static ERL_NIF_TERM get_decrypted_input_nif(ErlNifEnv *env, int argc,
         if (res <= 0) {
             if (SSL_get_error(state->ssl, res) != SSL_ERROR_WANT_READ) {
                 enif_mutex_unlock(state->mtx);
-		int reason = ERR_GET_REASON(ERR_peek_error());
-		if (reason == SSL_R_DATA_LENGTH_TOO_LONG ||
-		    reason == SSL_R_PACKET_LENGTH_TOO_LONG ||
-		    reason == SSL_R_UNKNOWN_PROTOCOL ||
-		    reason == SSL_R_UNEXPECTED_MESSAGE ||
-		    reason == SSL_R_WRONG_VERSION_NUMBER)
-		  /* Do not report badly formed Client Hello */
-		  return ERR_T(enif_make_atom(env, "closed"));
-		else if (state->sni_error)
-		  return ssl_error(env, state->sni_error);
-		else
-		  return ssl_error(env, "SSL_do_handshake failed");
+                int reason = ERR_GET_REASON(ERR_peek_error());
+                if (reason == SSL_R_DATA_LENGTH_TOO_LONG ||
+                    reason == SSL_R_PACKET_LENGTH_TOO_LONG ||
+                    reason == SSL_R_UNKNOWN_PROTOCOL ||
+                    reason == SSL_R_UNEXPECTED_MESSAGE ||
+                    reason == SSL_R_WRONG_VERSION_NUMBER)
+                    /* Do not report badly formed Client Hello */
+                    return ERR_T(enif_make_atom(env, "closed"));
+                else if (state->sni_error)
+                    return ssl_error(env, state->sni_error);
+                else
+                    return ssl_error(env, "SSL_do_handshake failed");
             }
         }
     }
@@ -1080,89 +1073,95 @@ static ERL_NIF_TERM get_decrypted_input_nif(ErlNifEnv *env, int argc,
 }
 
 static ERL_NIF_TERM add_certfile_nif(ErlNifEnv *env, int argc,
-				     const ERL_NIF_TERM argv[]) {
-  ErlNifBinary domain, file;
-  cert_info_t *info = NULL;
-  cert_info_t *old_info = NULL;
+                                     const ERL_NIF_TERM argv[]) {
+    ErlNifBinary domain, file;
+    cert_info_t *info = NULL;
+    cert_info_t *old_info = NULL;
 
-  if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
-    return enif_make_badarg(env);
-  if (!enif_inspect_iolist_as_binary(env, argv[1], &file))
-    return enif_make_badarg(env);
+    if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
+        return enif_make_badarg(env);
+    if (!enif_inspect_iolist_as_binary(env, argv[1], &file))
+        return enif_make_badarg(env);
 
-  info = enif_alloc(sizeof(cert_info_t));
-  if (info) {
-    memset(info, 0, sizeof(cert_info_t));
-    info->key = enif_alloc(domain.size+1);
-    info->file = enif_alloc(file.size+1);
-    if (info->key && info->file) {
-      memcpy(info->key, domain.data, domain.size);
-      memcpy(info->file, file.data, file.size);
-      info->key[domain.size] = 0;
-      info->file[file.size] = 0;
-      enif_rwlock_rwlock(certfiles_map_lock);
-      HASH_REPLACE_STR(certfiles_map, key, info, old_info);
-      free_cert_info(old_info);
-      enif_rwlock_rwunlock(certfiles_map_lock);
-    } else {
-      free_cert_info(info);
+    info = enif_alloc(sizeof(cert_info_t));
+    if (info) {
+        memset(info, 0, sizeof(cert_info_t));
+        info->key = enif_alloc(domain.size + 1);
+        info->file = enif_alloc(file.size + 1);
+        if (info->key && info->file) {
+            memcpy(info->key, domain.data, domain.size);
+            memcpy(info->file, file.data, file.size);
+            info->key[domain.size] = 0;
+            info->file[file.size] = 0;
+            enif_rwlock_rwlock(certfiles_map_lock);
+            HASH_REPLACE_STR(certfiles_map, key, info, old_info);
+            free_cert_info(old_info);
+            enif_rwlock_rwunlock(certfiles_map_lock);
+        } else {
+            free_cert_info(info);
+        }
     }
-  }
 
-  return enif_make_atom(env, "ok");
+    return enif_make_atom(env, "ok");
 }
 
 static ERL_NIF_TERM delete_certfile_nif(ErlNifEnv *env, int argc,
-					const ERL_NIF_TERM argv[]) {
-  ErlNifBinary domain;
-  char *ret = "false";
-  cert_info_t *info = NULL;
+                                        const ERL_NIF_TERM argv[]) {
+    ErlNifBinary domain;
+    char *ret = "false";
+    cert_info_t *info = NULL;
 
-  if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
-    return enif_make_badarg(env);
+    if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
+        return enif_make_badarg(env);
 
-  char key[domain.size+1];
-  memcpy(key, domain.data, domain.size);
-  key[domain.size] = 0;
-  enif_rwlock_rwlock(certfiles_map_lock);
-  HASH_FIND_STR(certfiles_map, key, info);
-  if (info) {
-    HASH_DEL(certfiles_map, info);
-    free_cert_info(info);
-    ret = "true";
-  }
-  enif_rwlock_rwunlock(certfiles_map_lock);
+    char key[domain.size + 1];
+    memcpy(key, domain.data, domain.size);
+    key[domain.size] = 0;
+    enif_rwlock_rwlock(certfiles_map_lock);
+    HASH_FIND_STR(certfiles_map, key, info);
+    if (info) {
+        HASH_DEL(certfiles_map, info);
+        free_cert_info(info);
+        ret = "true";
+    }
+    enif_rwlock_rwunlock(certfiles_map_lock);
 
-  return enif_make_atom(env, ret);
+    return enif_make_atom(env, ret);
 }
 
 static ERL_NIF_TERM get_certfile_nif(ErlNifEnv *env, int argc,
-				     const ERL_NIF_TERM argv[]) {
-  ErlNifBinary domain;
-  cert_info_t *info = NULL;
-  ERL_NIF_TERM file, result;
+                                     const ERL_NIF_TERM argv[]) {
+    ErlNifBinary domain;
+    cert_info_t *info = NULL;
+    ERL_NIF_TERM file, result;
 
-  if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
-    return enif_make_badarg(env);
+    if (!enif_inspect_iolist_as_binary(env, argv[0], &domain))
+        return enif_make_badarg(env);
 
-  char key[domain.size+1];
-  memcpy(key, domain.data, domain.size);
-  key[domain.size] = 0;
-  enif_rwlock_rlock(certfiles_map_lock);
-  info = lookup_certfile(key);
-  if (info) {
-    unsigned char *tmp = enif_make_new_binary(env, strlen(info->file), &file);
-    if (tmp) {
-      memcpy(tmp, info->file, strlen(info->file));
-      result = enif_make_tuple2(env, enif_make_atom(env, "ok"), file);
-    } else
-      result = enif_make_atom(env, "error");
-  } else {
-    result = enif_make_atom(env, "error");
-  }
-  enif_rwlock_runlock(certfiles_map_lock);
+    char key[domain.size + 1];
+    memcpy(key, domain.data, domain.size);
+    key[domain.size] = 0;
+    enif_rwlock_rlock(certfiles_map_lock);
+    info = lookup_certfile(key);
+    if (info) {
+        unsigned char *tmp = enif_make_new_binary(env, strlen(info->file), &file);
+        if (tmp) {
+            memcpy(tmp, info->file, strlen(info->file));
+            result = enif_make_tuple2(env, enif_make_atom(env, "ok"), file);
+        } else
+            result = enif_make_atom(env, "error");
+    } else {
+        result = enif_make_atom(env, "error");
+    }
+    enif_rwlock_runlock(certfiles_map_lock);
 
-  return result;
+    return result;
+}
+
+static ERL_NIF_TERM clear_cache_nif(ErlNifEnv *env, int argc,
+                                    const ERL_NIF_TERM argv[]) {
+    clear_certs_map();
+    return enif_make_atom(env, "ok");
 }
 
 static ERL_NIF_TERM invalidate_nif(ErlNifEnv *env, int argc,
@@ -1184,19 +1183,56 @@ static ERL_NIF_TERM invalidate_nif(ErlNifEnv *env, int argc,
     return enif_make_atom(env, "ok");
 }
 
+static ERL_NIF_TERM get_negotiated_cipher_nif(ErlNifEnv *env, int argc,
+                                              const ERL_NIF_TERM argv[]) {
+    state_t *state = NULL;
+
+    if (argc != 1)
+        return enif_make_badarg(env);
+
+    if (!enif_get_resource(env, argv[0], tls_state_t, (void *) &state))
+        return enif_make_badarg(env);
+
+    if (!state->mtx || !state->ssl) return enif_make_badarg(env);
+
+    enif_mutex_lock(state->mtx);
+
+    if (!state->valid) {
+        enif_mutex_unlock(state->mtx);
+        return ERR_T(enif_make_atom(env, "closed"));
+    }
+
+    const char *version = SSL_get_version(state->ssl);
+    const char *cipher = SSL_get_cipher_name(state->ssl);
+    enif_mutex_unlock(state->mtx);
+
+    ErlNifBinary bin;
+    size_t vl = strlen(version);
+    size_t cl = strlen(cipher);
+    if (!enif_alloc_binary(vl + cl + 1, &bin)) {
+        return ERR_T(enif_make_atom(env, "enomem"));
+    }
+    memcpy(bin.data, version, vl);
+    bin.data[vl] = ' ';
+    memcpy(bin.data + vl + 1, cipher, cl);
+    return enif_make_binary(env, &bin);
+}
+
 static ErlNifFunc nif_funcs[] =
         {
-                {"open_nif",                 8, open_nif},
-                {"set_encrypted_input_nif",  2, set_encrypted_input_nif},
-                {"set_decrypted_output_nif", 2, set_decrypted_output_nif},
-                {"get_decrypted_input_nif",  2, get_decrypted_input_nif},
-                {"get_encrypted_output_nif", 1, get_encrypted_output_nif},
-                {"get_verify_result_nif",    1, get_verify_result_nif},
-                {"get_peer_certificate_nif", 1, get_peer_certificate_nif},
-		{"add_certfile_nif",         2, add_certfile_nif},
-		{"delete_certfile_nif",      1, delete_certfile_nif},
-		{"get_certfile_nif",         1, get_certfile_nif},
-                {"invalidate_nif",           1, invalidate_nif}
+                {"open_nif",                  8, open_nif},
+                {"set_encrypted_input_nif",   2, set_encrypted_input_nif},
+                {"set_decrypted_output_nif",  2, set_decrypted_output_nif},
+                {"get_decrypted_input_nif",   2, get_decrypted_input_nif},
+                {"get_encrypted_output_nif",  1, get_encrypted_output_nif},
+                {"get_verify_result_nif",     1, get_verify_result_nif},
+                {"get_peer_certificate_nif",  1, get_peer_certificate_nif},
+                {"add_certfile_nif",          2, add_certfile_nif},
+                {"delete_certfile_nif",       1, delete_certfile_nif},
+                {"get_certfile_nif",          1, get_certfile_nif},
+                {"clear_cache_nif",           0, clear_cache_nif},
+                {"invalidate_nif",            1, invalidate_nif},
+                {"get_negotiated_cipher_nif", 1, get_negotiated_cipher_nif}
         };
 
 ERL_NIF_INIT(fast_tls, nif_funcs, load, NULL, NULL, unload)
